(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var Accounts = Package['accounts-base'].Accounts;
var SandstormDb = Package['sandstorm-db'].SandstormDb;
var SandstormPermissions = Package['sandstorm-permissions'].SandstormPermissions;

/* Package-scope variables */
var SandstormBackend, inMeteor, promiseToFuture, waitPromise, SANDSTORM_ALTHOME;

(function () {

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// packages/sandstorm-backend/sandstorm-backend.js                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
                                                                                                                       // 16
var Capnp = Npm.require("capnp");                                                                                      // 17
var Backend = Capnp.importSystem("sandstorm/backend.capnp").Backend;                                                   // 18
var Crypto = Npm.require("crypto");                                                                                    // 19
var Future = Npm.require("fibers/future");                                                                             // 20
var Promise = Npm.require("es6-promise").Promise;                                                                      // 21
                                                                                                                       // 22
var inMeteorInternal = Meteor.bindEnvironment(function (callback) {                                                    // 23
  callback();                                                                                                          // 24
});                                                                                                                    // 25
                                                                                                                       // 26
inMeteor = function (callback) {                                                                                       // 27
  // Calls the callback in a Meteor context.  Returns a Promise for its result.                                        // 28
  return new Promise(function (resolve, reject) {                                                                      // 29
    inMeteorInternal(function () {                                                                                     // 30
      try {                                                                                                            // 31
        resolve(callback());                                                                                           // 32
      } catch (err) {                                                                                                  // 33
        reject(err);                                                                                                   // 34
      }                                                                                                                // 35
    });                                                                                                                // 36
  });                                                                                                                  // 37
}                                                                                                                      // 38
                                                                                                                       // 39
promiseToFuture = function (promise) {                                                                                 // 40
  var result = new Future();                                                                                           // 41
  promise.then(result.return.bind(result), result.throw.bind(result));                                                 // 42
  return result;                                                                                                       // 43
}                                                                                                                      // 44
                                                                                                                       // 45
waitPromise = function (promise) {                                                                                     // 46
  return promiseToFuture(promise).wait();                                                                              // 47
}                                                                                                                      // 48
                                                                                                                       // 49
SANDSTORM_ALTHOME = Meteor.settings && Meteor.settings.home;                                                           // 50
                                                                                                                       // 51
SandstormBackend = function(db, backendCap) {                                                                          // 52
  this._db = db;                                                                                                       // 53
  this.runningGrains = {};                                                                                             // 54
  this._backendCap = backendCap                                                                                        // 55
}                                                                                                                      // 56
                                                                                                                       // 57
SandstormBackend.prototype.cap = function() {                                                                          // 58
  return this._backendCap;                                                                                             // 59
}                                                                                                                      // 60
                                                                                                                       // 61
SandstormBackend.prototype.shutdownGrain = function (grainId, ownerId, keepSessions) {                                 // 62
  if (!keepSessions) {                                                                                                 // 63
    Sessions.remove({grainId: grainId});                                                                               // 64
    delete this.runningGrains[grainId];                                                                                // 65
  }                                                                                                                    // 66
                                                                                                                       // 67
  var grain = this._backendCap.getGrain(ownerId, grainId).supervisor;                                                  // 68
  return grain.shutdown().then(function () {                                                                           // 69
    grain.close();                                                                                                     // 70
    throw new Error("expected shutdown() to throw disconnected");                                                      // 71
  }, function (err) {                                                                                                  // 72
    grain.close();                                                                                                     // 73
    if (err.kjType !== "disconnected") {                                                                               // 74
      throw err;                                                                                                       // 75
    }                                                                                                                  // 76
  });                                                                                                                  // 77
}                                                                                                                      // 78
                                                                                                                       // 79
SandstormBackend.prototype.deleteGrain = function (grainId, ownerId) {                                                 // 80
  // We leave it up to the caller if they want to actually wait, but some don't so we report                           // 81
  // exceptions.                                                                                                       // 82
  return this._backendCap.deleteGrain(ownerId, grainId).catch(function (err) {                                         // 83
    console.error("problem deleting grain " + grainId + ":", err.message);                                             // 84
    throw err;                                                                                                         // 85
  });                                                                                                                  // 86
}                                                                                                                      // 87
                                                                                                                       // 88
                                                                                                                       // 89
SandstormBackend.prototype.openGrain = function (grainId, isRetry) {                                                   // 90
  // Create a Cap'n Proto connection to the given grain. Note that this function does not actually                     // 91
  // verify that the connection succeeded. Instead, if an RPC call to the connection fails, check                      // 92
  // shouldRestartGrain(). If it returns true, call continueGrain() and then openGrain()                               // 93
  // again with isRetry = true, and then retry.                                                                        // 94
  //                                                                                                                   // 95
  // Must be called in a Meteor context.                                                                               // 96
                                                                                                                       // 97
  if (isRetry) {                                                                                                       // 98
    // Since this is a retry, try starting the grain even if we think it's already running.                            // 99
    return this.continueGrain(grainId);                                                                                // 100
  } else {                                                                                                             // 101
    // Start the grain if it is not running.                                                                           // 102
    var runningGrain = this.runningGrains[grainId];                                                                    // 103
    if (runningGrain) {                                                                                                // 104
      return waitPromise(runningGrain);                                                                                // 105
    } else {                                                                                                           // 106
      return this.continueGrain(grainId);                                                                              // 107
    }                                                                                                                  // 108
  }                                                                                                                    // 109
}                                                                                                                      // 110
                                                                                                                       // 111
SandstormBackend.shouldRestartGrain = function (error, retryCount) {                                                   // 112
  // Given an error thrown by an RPC call to a grain, return whether or not it makes sense to try                      // 113
  // to restart the grain and retry. `retryCount` is the number of times that the request has                          // 114
  // already gone through this cycle (should be zero for the first call).                                              // 115
                                                                                                                       // 116
  return error.kjType === "disconnected" && retryCount < 1;                                                            // 117
}                                                                                                                      // 118
                                                                                                                       // 119
SandstormBackend.prototype.maybeRetryUseGrain = function (grainId, cb, retryCount, err) {                              // 120
  var self = this;                                                                                                     // 121
  if (SandstormBackend.shouldRestartGrain(err, retryCount)) {                                                          // 122
    return inMeteor(function () {                                                                                      // 123
      return cb(self.openGrain(grainId, true).supervisor)                                                              // 124
          .catch(self.maybeRetryUseGrain.bind(self, grainId, cb, retryCount + 1));                                     // 125
    });                                                                                                                // 126
  } else {                                                                                                             // 127
    throw err;                                                                                                         // 128
  }                                                                                                                    // 129
}                                                                                                                      // 130
                                                                                                                       // 131
SandstormBackend.prototype.useGrain = function (grainId, cb) {                                                         // 132
  // This will open a grain for you, handling restarts if needed, and call the passed function with                    // 133
  // the supervisor capability as the only parameter. The callback must return a promise that used                     // 134
  // the supervisor, so that we can check if a disconnect error occurred, and retry if possible.                       // 135
  // This function returns the same promise that your callback returns.                                                // 136
  //                                                                                                                   // 137
  // This function is NOT expected to be run in a meteor context.                                                      // 138
                                                                                                                       // 139
  var runningGrain = this.runningGrains[grainId];                                                                      // 140
  var self = this;                                                                                                     // 141
  if (runningGrain) {                                                                                                  // 142
    return runningGrain.then(function (grainInfo) {                                                                    // 143
      return cb(grainInfo.supervisor);                                                                                 // 144
    }).catch(self.maybeRetryUseGrain.bind(self, grainId, cb, 0));                                                      // 145
  } else {                                                                                                             // 146
    return inMeteor(function () {                                                                                      // 147
      return cb(self.openGrain(grainId, false).supervisor)                                                             // 148
          .catch(self.maybeRetryUseGrain.bind(self, grainId, cb, 0));                                                  // 149
    });                                                                                                                // 150
  }                                                                                                                    // 151
}                                                                                                                      // 152
                                                                                                                       // 153
SandstormBackend.prototype.continueGrain = function(grainId) {                                                         // 154
  var grain = Grains.findOne(grainId);                                                                                 // 155
  if (!grain) {                                                                                                        // 156
    throw new Meteor.Error(404, "Grain Not Found", "Grain ID: " + grainId);                                            // 157
  }                                                                                                                    // 158
                                                                                                                       // 159
  // If a DevPackage with the same app ID is currently active, we let it override the installed                        // 160
  // package, so that the grain runs using the dev app.                                                                // 161
  var devPackage = DevPackages.findOne({appId: grain.appId});                                                          // 162
  var isDev;                                                                                                           // 163
  var pkg;                                                                                                             // 164
  if (devPackage) {                                                                                                    // 165
    isDev = true;                                                                                                      // 166
    pkg = devPackage;                                                                                                  // 167
  } else {                                                                                                             // 168
    pkg = Packages.findOne(grain.packageId);                                                                           // 169
    if (!pkg) {                                                                                                        // 170
      throw new Meteor.Error(500, "Grain's package not installed",                                                     // 171
                             "Package ID: " + grain.packageId);                                                        // 172
    }                                                                                                                  // 173
  }                                                                                                                    // 174
  var manifest = pkg.manifest;                                                                                         // 175
  var packageId = pkg._id;                                                                                             // 176
                                                                                                                       // 177
  if (!("continueCommand" in manifest)) {                                                                              // 178
    throw new Meteor.Error(500, "Package manifest defines no continueCommand.",                                        // 179
                           "Package ID: " + packageId);                                                                // 180
  }                                                                                                                    // 181
                                                                                                                       // 182
  return this.startGrainInternal(                                                                                      // 183
      packageId, grainId, grain.userId, manifest.continueCommand, false, isDev);                                       // 184
}                                                                                                                      // 185
                                                                                                                       // 186
SandstormBackend.prototype.startGrainInternal = function(packageId, grainId, ownerId, command, isNew, isDev) {         // 187
  // Starts the grain supervisor.  Must be executed in a Meteor context.  Blocks until grain is                        // 188
  // started. Returns a promise for an object containing two fields: `owner` (the ID of the owning                     // 189
  // user) and `supervisor` (the supervisor capability).                                                               // 190
                                                                                                                       // 191
  if (isUserExcessivelyOverQuota(Meteor.users.findOne(ownerId))) {                                                     // 192
    throw new Meteor.Error(402, "Cannot start grain because owner's storage is exhausted.");                           // 193
  }                                                                                                                    // 194
                                                                                                                       // 195
  // Ugly: Stay backwards-compatible with old manifests that had "executablePath" and "args" rather                    // 196
  //   than just "argv".                                                                                               // 197
  if ("args" in command) {                                                                                             // 198
    if (!("argv" in command)) {                                                                                        // 199
      command.argv = command.args;                                                                                     // 200
    }                                                                                                                  // 201
    delete command.args;                                                                                               // 202
  }                                                                                                                    // 203
  if ("executablePath" in command) {                                                                                   // 204
    if (!("deprecatedExecutablePath" in command)) {                                                                    // 205
      command.deprecatedExecutablePath = command.executablePath;                                                       // 206
    }                                                                                                                  // 207
    delete command.executablePath;                                                                                     // 208
  }                                                                                                                    // 209
                                                                                                                       // 210
  var whenReady = this._backendCap.startGrain(ownerId, grainId, packageId, command, isNew, isDev)                      // 211
      .then(function (results) {                                                                                       // 212
    return {                                                                                                           // 213
      owner: ownerId,                                                                                                  // 214
      supervisor: results.supervisor                                                                                   // 215
    };                                                                                                                 // 216
  });                                                                                                                  // 217
                                                                                                                       // 218
  this.runningGrains[grainId] = whenReady;                                                                             // 219
  return waitPromise(whenReady);                                                                                       // 220
}                                                                                                                      // 221
                                                                                                                       // 222
SandstormBackend.prototype.updateLastActive = function(grainId, userId, identityId) {                                  // 223
  // Update the lastActive date on the grain, any relevant API tokens, and the user,                                   // 224
  // and also update the user's storage usage.                                                                         // 225
                                                                                                                       // 226
  var storagePromise = undefined;                                                                                      // 227
  if (Meteor.settings.public.quotaEnabled) {                                                                           // 228
    storagePromise = globalBackend._backendCap.getUserStorageUsage(userId);                                            // 229
  }                                                                                                                    // 230
                                                                                                                       // 231
  var now = new Date();                                                                                                // 232
  if (Grains.update(grainId, {$set: {lastUsed: now}}) === 0) {                                                         // 233
    // Grain must have been deleted. Ignore.                                                                           // 234
    return;                                                                                                            // 235
  }                                                                                                                    // 236
  if (userId) {                                                                                                        // 237
    Meteor.users.update(userId, {$set: {lastActive: now}});                                                            // 238
  }                                                                                                                    // 239
  if (identityId) {                                                                                                    // 240
    // Update any API tokens that match this user/grain pairing as well                                                // 241
    var now = new Date();                                                                                              // 242
    ApiTokens.update({"grainId": grainId, "owner.user.identityId": identityId},                                        // 243
        {$set: {"owner.user.lastUsed": now }});                                                                        // 244
  }                                                                                                                    // 245
                                                                                                                       // 246
  if (Meteor.settings.public.quotaEnabled) {                                                                           // 247
    try {                                                                                                              // 248
      var ownerId = Grains.findOne(grainId).userId;                                                                    // 249
      var size = parseInt(waitPromise(storagePromise).size);                                                           // 250
      Meteor.users.update(ownerId, {$set: {storageUsage: size}});                                                      // 251
      // TODO(security): Consider actively killing grains if the user is excessively over quota?                       // 252
      //   Otherwise a constantly-active grain could consume arbitrary space without being stopped.                    // 253
    } catch (err) {                                                                                                    // 254
      if (err.kjType !== "unimplemented") {                                                                            // 255
        console.error("error getting user storage usage:", err.stack);                                                 // 256
      }                                                                                                                // 257
    }                                                                                                                  // 258
  }                                                                                                                    // 259
}                                                                                                                      // 260
                                                                                                                       // 261
                                                                                                                       // 262
function generateSessionId(grainId, userId, salt) {                                                                    // 263
  var sessionParts = [grainId, salt];                                                                                  // 264
  if (userId) {                                                                                                        // 265
    sessionParts.push(userId);                                                                                         // 266
  }                                                                                                                    // 267
  var sessionInput = sessionParts.join(":");                                                                           // 268
  return Crypto.createHash("sha256").update(sessionInput).digest("hex");                                               // 269
}                                                                                                                      // 270
                                                                                                                       // 271
SandstormBackend.prototype.openSessionInternal = function (grainId, userId, identityId, title, apiToken, cachedSalt) { // 272
                                                                                                                       // 273
  // Start the grain if it is not running. This is an optimization: if we didn't start it here,                        // 274
  // it would start on the first request to the session host, but we'd like to get started before                      // 275
  // the round trip.                                                                                                   // 276
  var runningGrain = this.runningGrains[grainId];                                                                      // 277
  var grainInfo;                                                                                                       // 278
  if (runningGrain) {                                                                                                  // 279
    grainInfo = waitPromise(runningGrain);                                                                             // 280
  } else {                                                                                                             // 281
    grainInfo = this.continueGrain(grainId);                                                                           // 282
  }                                                                                                                    // 283
                                                                                                                       // 284
  this.updateLastActive(grainId, userId, identityId);                                                                  // 285
                                                                                                                       // 286
  cachedSalt = cachedSalt || Random.id(22);                                                                            // 287
  var sessionId = generateSessionId(grainId, userId, cachedSalt);                                                      // 288
  var session = Sessions.findOne({_id: sessionId});                                                                    // 289
  if (session) {                                                                                                       // 290
    // TODO(someday): also do some more checks for anonymous sessions (sessions without a userId).                     // 291
    if ((session.identityId && session.identityId !== identityId) ||                                                   // 292
        (session.grainId !== grainId)) {                                                                               // 293
      var e = new Meteor.Error(500, "Duplicate SessionId");                                                            // 294
      console.error(e);                                                                                                // 295
      throw e;                                                                                                         // 296
    } else {                                                                                                           // 297
      return {sessionId: session._id, title: title, grainId: grainId, hostId: session.hostId, salt: cachedSalt};       // 298
    }                                                                                                                  // 299
  }                                                                                                                    // 300
                                                                                                                       // 301
  session = {                                                                                                          // 302
    _id: sessionId,                                                                                                    // 303
    grainId: grainId,                                                                                                  // 304
    hostId: Crypto.createHash("sha256").update(sessionId).digest("hex").slice(0, 32),                                  // 305
    timestamp: new Date().getTime(),                                                                                   // 306
    hasLoaded: false                                                                                                   // 307
  };                                                                                                                   // 308
                                                                                                                       // 309
  if (userId) {                                                                                                        // 310
    session.identityId = identityId;                                                                                   // 311
    session.userId = userId;                                                                                           // 312
  } else if (apiToken) {                                                                                               // 313
    session.hashedToken = apiToken._id;                                                                                // 314
  } else {                                                                                                             // 315
    // Must be old-style sharing, i.e. !grain.private.                                                                 // 316
  }                                                                                                                    // 317
                                                                                                                       // 318
  Sessions.insert(session);                                                                                            // 319
                                                                                                                       // 320
  return {sessionId: session._id, title: title, grainId: grainId, hostId: session.hostId, salt: cachedSalt};           // 321
}                                                                                                                      // 322
                                                                                                                       // 323
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-backend'] = {
  SandstormBackend: SandstormBackend
};

})();

//# sourceMappingURL=sandstorm-backend.js.map
